#pragma once
#ifndef DOMAIN_HPP
#define DOMAIN_HPP 1
#include <field.hpp>
#include <vector>
#include <beam.hpp>
#include <triangle.hpp>
#include <vsrc.hpp>
#include <horn.hpp>
#include <grid.hpp>

using namespace QOSM::EM;

namespace qpropa = QOSM::Propagation;
namespace qitem = QOSM::Item;
namespace qem = QOSM::EM;
namespace qpose = QOSM::Pose;
namespace qsrc = QOSM::EM::Source;

namespace QOSM::Propagation{
    enum class Dim{
        x = 0,
        y,
    };

    /**
     * @brief
     * Class modelling an domain of computation (GBS + GBE + GBT)
     *
     */
    struct Domain{
        std::vector<qitem::triangle_ptr> geometry;
        
        VectorNx3 src_r_pts;
        
        double w0;
        double threshPw;
        std::array<int, 2> numBouces{0, 0};
        bool do_gbe = false;
        bool do_gbt = false;
        int dim = -1;

        /**
         * Create a domain of computation
         * @param geometry List of all triangle of all objects contained by the domain
         * @param dim index of the dimention to use. -1 enables the 3D modelling, 0 or 1 enables the 2D modelling
         */
        Domain( 
            const std::vector<qitem::Triangle> &triangles,
            const int dim = -1
        );

        /**
         * Set dimension
         * @param dim index of the dimention to use. -1 enables the 3D modelling, 0 or 1 enables the 2D modelling
         */
        void setDimension(const int dim);
        
        /**
         * Set GBT settings
         * @param numBouces [2 integers] Maximum bources allowed for a beam for (reflection, transmission)
         * @param pw_threshold Minimal power ratio (to the incident beam power) necessary for generated beams to be accounted.
         */
        void setGBTsettings(
            const std::array<int, 2> &numBouces,
            const double &threshPw = 1e-7
        );

        /**
         * Perform the field propagation through the quasi-optical (sub-)system
         * @param source Electromagnetic source (Horn, VirtualSource)
         * @return a list of Beams
         */
        qsrc::VirtualSource propagates(
            const std::shared_ptr<qsrc::Source> &source,
            const QOSM::Tools::Grid &gbeGrig,
            const double &samplingStep,
            const double &kappa
        );

        /**
         * Perform the field propagation through the quasi-optical (sub-)system
         * @param source Electromagnetic source (Horn, VirtualSource)
         * @return a list of Beams
         */
        qsrc::VirtualSource propagates(
            const std::shared_ptr<qsrc::Source> &source,
            const VectorNx3 &src_r_gbe,
            const double &samplingStep,
            const double &kappa
        );
        
        qsrc::VirtualSource propagates(
            const qsrc::VirtualSource &source
        );
        
    };
}
#endif