#pragma once
#ifndef FRAME_HPP
#define FRAME_HPP 1
#define NUMCPP_NO_USE_BOOST ON

#include "pose.hpp"
#include "field.hpp"
#include <array>

namespace QOSM::Pose{
    struct Frame{
        /// [m] Origin of the local frame w.r.t. the origin of the reference frame
        Vec3 ori; 

        /// Axes of the local frame written in the reference frame
        Vec3 x, y, z; 

        /// rotation matrix from the local frame to the reference frame as [R11, R12, R13, R21, R22, R23, R31, R32, R33]
        std::array<double, 9> R;

        /**
         * Create a Frame f_loc_ref with the base (x, y, z) and the origin vector
         * @param x_loc_ref x-axis of the frame (loc) written in the reference frame (ref)
         * @param y_loc_ref y-axis of the frame (loc) written in the reference frame (ref)
         * @param z_loc_ref z-axis of the frame (loc) written in the reference frame (ref)
         * @param ref_r_ref_loc center of the frame (loc) written in the reference frame (ref)
        */
        Frame(
            Vec3 x_loc_ref = Vec3(1., 0, 0), 
            Vec3 y_loc_ref = Vec3(0, 1., 0), 
            Vec3 z_loc_ref = Vec3(0, 0, 1.), 
            Vec3 ref_r_ref_loc = Vec3()
        );

        /**
         * Create a Frame f_loc_ref using a pose (either a quaternion or a rotation vector)
         * @param values (x, y, z, rx, ry, rz) in [m], [m.rad] or (x, y, z, qw, qx, qy, qz)
        */
        Frame(std::initializer_list<double> values);

        /**
         * Create a Frame f_loc_ref using a pose (either a quaternion or a rotation vector)
         * @param ref_r_ref_loc center of the local frame (loc) written in the reference frame (ref)
         * @param q_ref_loc quaternion from loc to ref
        */
        Frame(const Vec3 &ref_r_ref_loc, const Quaternion &q_ref_loc);

        /**
         * Create a Frame f_loc_ref using a pose object
         * @param rpose
        */
        Frame(const Pose &rpose);

        /**
         * Create a Frame f_loc_ref using an axis-angle
         * @param ref_r_ref_loc center of the frame (loc) written in the reference frame (ref)
         * @param u_ref axis of the local frame (loc) rotation written in the reference frame (ref)
         * @param angle angle of rotation
         * @param deg True of angle is set in degrees, otherwise expects radian
        */
        Frame(const Vec3 &ref_r_ref_loc, const Vec3 &u_ref, const double &angle, const bool &deg = false) ;

        /**
         * Write a position vector written in (ref) frame to (loc) frame
         * @param ref_r_ref_pt position vector written in the reference frame (ref)
         * @return loc_r_loc_pt position vector written in the local frame (loc)
        */
        Vec3 toLocal(const Vec3 &ref_r_ref_pt) const;         
        
        /**
         * Write a position vector written in (loc) frame to (ref) frame
         * @param loc_r_loc_pt position vector written in the local frame (ref)
         * @return ref_r_ref_pt position vector written in the reference frame (loc)
        */       
        Vec3 toRef(const Vec3 &loc_r_loc_pt) const;

        /**
         * Write a unitary vector written in (ref) frame to (loc) frame
         * @param u_ref position vector written in the reference frame (ref)
         * @return u_loc position vector written in the local frame (loc)
        */
        Vec3 rot2Local(const Vec3 &u_ref) const;   
        
        /**
         * Write a unitary vector written in (loc) frame to (ref) frame
         * @param u_loc position vector written in the local frame (loc)
         * @return u_ref position vector written in the reference frame (ref)
        */  
        Vec3 rot2Ref(const Vec3 &u_loc) const;

        /**
         * Write a bunch of position vectors written in (ref) frame to (loc) frame
         * @param ref_r_ref_pt position vectors written in the reference frame (ref)
         * @return loc_r_loc_pt position vectors written in the local frame (loc)
        */  
        VectorNx3 toLocal(const VectorNx3 &ref_r_ref_pt) const;
            
        /**
         * Write a bunch of position vectors written in (loc) frame to (ref) frame
         * @param loc_r_loc_pt position vectors written in the local frame (ref)
         * @return ref_r_ref_pt position vectors written in the reference frame (loc)
        */  
        VectorNx3 toRef(const VectorNx3 &loc_r_loc_pt) const;
        
        /**
         * Rotate a bunch of position vectors written in (ref) frame to (loc) frame
         * @param v_pt_ref position vectors written in the reference frame (ref)
         * @return v_pt_loc position vectors written in the local frame (loc)
        */ 
        VectorNx3 rot2Local(const VectorNx3 &v_pt_ref) const;    
        
        /**
         * Rotate a bunch of position vectors written in (loc) frame to (ref) frame
         * @param v_pt_loc position vectors written in the local frame (ref)
         * @return v_pt_ref position vectors written in the reference frame (loc)
        */  
        VectorNx3 rot2Ref(const VectorNx3 &v_pt_loc) const;

        /**
         * Rotate a complex vector written in (ref) frame to (loc) frame
         * @param v_ref position vector written in the reference frame (ref)
         * @return v_loc position vector written in the local frame (loc)
        */ 
        std::array<std::complex<double>, 3> rot2Local(const std::array<std::complex<double>, 3> &v_ref) const;
            
        /**
         * Rotate a complex vector written in (loc) frame to (ref) frame
         * @param v_loc position vector written in the local frame (loc)
         * @return v_ref position vector written in the reference frame (ref)
        */ 
        std::array<std::complex<double>, 3> rot2Ref(const std::array<std::complex<double>, 3> &v_pt_loc) const;

        /**
         * Write a complex field vector written in (ref) frame to (loc) frame
         * @param ref_E_pt field vector written in the reference frame (ref)
         * @return loc_E_pt field vector written in the local frame (loc)
        */ 
        EM::FieldNx3 toLocal(const EM::FieldNx3 &ref_E_pt) const;
            
        /**
         * Rotate a complex field vector written in (loc) frame to (ref) frame
         * @param loc_E_pt field vector written in the local frame (loc)
         * @return ref_E_pt field vector written in the reference frame (ref)
        */    
        EM::FieldNx3 toRef(const EM::FieldNx3 &loc_E_pt) const;

        friend std::ostream& operator<<(std::ostream& os, const Frame& o){
            os << "Frame: " << std::endl 
                << "   ori: " << o.ori << std::endl
                << "   x: " << o.x << std::endl
                << "   y: " << o.y << std::endl
                << "   z: " << o.z << std::endl;
            return os;
        }
    };
}
#endif