#include <domain.hpp>
#include <propagation.hpp>
#include <memory>

namespace qem = QOSM::EM;
namespace qpropa = QOSM::Propagation;
namespace qitem = QOSM::Item;
namespace qpose = QOSM::Pose;
namespace qsrc = QOSM::EM::Source;

using namespace QOSM::Propagation;

Domain::Domain( 
    const std::vector<qitem::Triangle> &triangles,
    const int dim
) : threshPw(1e-7), do_gbe(false), do_gbt(false), dim(dim){
    for (const qitem::Triangle &triangle : triangles){
        geometry.push_back(std::make_shared<qitem::Triangle>(std::move(triangle)));
    }
}

void Domain::setDimension(int dim){
    if (dim > 1 || dim <-1){
        throw std::runtime_error("Invalid dimentions, must be -1, 0 or 1");
    }
    this->dim = dim;
}

void Domain::setGBTsettings(
    const std::array<int, 2> &numBouces,
    const double &threshPw
){
    this->numBouces = numBouces;
    this->threshPw = threshPw;
    do_gbt = true;
}

qsrc::VirtualSource Domain::propagates(
    const std::shared_ptr<qsrc::Source> &source,
    const VectorNx3 &src_r_gbe,
    const double &samplingStep,
    const double &kappa
){
    qsrc::VirtualSource vsrcRes = qsrc::VirtualSource(source->freq * 1e-9);

    const auto res = source->fields(src_r_gbe);
    const auto &S = std::get<2>(res);
    
    std::vector<Beam> beams;
    if (dim < 0){
        const auto &E = std::get<0>(res);
        beams = fieldExpansion(E, S, src_r_gbe, source->freq, samplingStep * kappa, threshPw);
    } else {
        const auto &E = std::get<0>(res);
        const auto Es = FieldNx1(E, dim);
        beams = fieldExpansion(Es, S, src_r_gbe, source->freq, dim, samplingStep * kappa, 0.);
        vsrcRes.init(beams);
        vsrcRes.setEzIs0Assumption(true);
    }

    if (do_gbt){
        auto beams_gbt_q = beamTracing(geometry, beams, numBouces, threshPw, false);

        // convert from deque to vector
        std::vector<Beam> beams_gbt = {beams_gbt_q.begin(), beams_gbt_q.end()};

        // set the vsrc with the new set of beams
        vsrcRes.init(beams_gbt);
    }

    return vsrcRes;
}

qsrc::VirtualSource Domain::propagates(
    const std::shared_ptr<qsrc::Source> &source,
    const QOSM::Tools::Grid &gbeGrig,
    const double &samplingStep, 
    const double &kappa
){
    const auto src_r_gbe = gbeGrig.points;
    return propagates(source, gbeGrig.points, samplingStep, kappa);
}

qsrc::VirtualSource Domain::propagates(
    const qsrc::VirtualSource &source
){
    qsrc::VirtualSource vsrcRes = qsrc::VirtualSource(source.freq * 1e-9);
    auto beams = source.beams;
    auto beams_gbt_q = beamTracing(geometry, beams, numBouces, threshPw, false);

    // convert from deque to vector
    std::vector<Beam> beams_gbt = {beams_gbt_q.begin(), beams_gbt_q.end()};

    // set the vsrc with the new set of beams
    vsrcRes.init(beams_gbt);

    return vsrcRes;
}
