#include <grid.hpp>
#include <omp.h>

using namespace QOSM::Tools;
using namespace QOSM::Pose;

void Grid::createPoints(const std::array<double,2> &uRange, const std::array<double,2> &vRange, const std::array<int,2> &numPoints,  const double &n, const PlaneType &plane){
    u = nc::linspace(uRange[0], uRange[1], numPoints[0], true).reshape(numPoints[0]);
    v = nc::linspace(vRange[0], vRange[1], numPoints[1], true).reshape(numPoints[1]);
    const auto grid = nc::meshgrid(u, v);
    auto U = grid.first.flatten();
    auto V = grid.second.flatten();

    points = VectorNx3(U.size(), n);
    const auto r = points.rSlice();
    switch (plane){
        case PlaneType::XY:
            points.putCol(0, U);
            points.putCol(1, V);
            shape = {u.size(), v.size()};
            break;

        case PlaneType::XZ:
            points.putCol(0, U);
            points.putCol(2, V);
            shape = {u.size(), v.size()};
            break;

        case PlaneType::YX:
            points.putCol(0, V);
            points.putCol(1, U);
            shape = {v.size(), u.size()};
            break;

        case PlaneType::YZ:
            points.putCol(1, U);
            points.putCol(2, V);
            shape = {u.size(), v.size()};
            break;

        case PlaneType::ZX:
            points.putCol(0, V);
            points.putCol(2, U);
            shape = {v.size(), u.size()};
            break;

        case PlaneType::ZY:
            points.putCol(1, V);
            points.putCol(2, U);
            shape = {v.size(), u.size()};
            break;
    
        default:
            points = VectorNx3(0);
            break;
    }
}

Grid::Grid(const std::array<double,3> &uRange, const std::array<double,3> &vRange, const double &n, const PlaneType &plane){
    const double ru = (uRange[1] - uRange[0]) / 2;
    const double rv = (vRange[1] - vRange[0]) / 2;
    const int num_u = 1 + int(2 * ru / uRange[2]);
    const int num_v = 1 + int(2 * rv / vRange[2]);

    createPoints({uRange[0], uRange[1]}, {vRange[0], vRange[1]}, {num_u, num_v}, n, plane);
}


Grid::Grid(const std::array<double,2> &uRange, const std::array<double,2> &vRange, const std::array<int,2> &numPoints, const double &n, const PlaneType &plane){
    createPoints(uRange, vRange, numPoints, n, plane);
}

void Grid::rotateAlongU(const Vec3 &centre, const double &angle, const bool &deg){
    const double x = (double)(type==PlaneType::XY || type==PlaneType::XZ);
    const double y = (double)(type==PlaneType::YX || type==PlaneType::YZ);
    const double z = (double)(type==PlaneType::ZX || type==PlaneType::ZY);
    const auto rot = Quaternion(Vec3(x, y, z), angle, deg);
    rotate(centre, rot);
}

void Grid::rotateAlongV(const Vec3 &centre, const double &angle, const bool &deg){
    const double x = (double)(type==PlaneType::YX || type==PlaneType::ZX);
    const double y = (double)(type==PlaneType::XY || type==PlaneType::ZY);
    const double z = (double)(type==PlaneType::XZ || type==PlaneType::YZ);
    const auto rot = Quaternion(Vec3(x, y, z), angle, deg);
    rotate(centre, rot);
}


void Grid::rotate(const Vec3 &centre, const Quaternion &rot){
    #pragma omp parallel for
    for (int i=0; i<(int)points.shape().rows; i++){
        Vec3 row(points[i, 0], points[i, 1], points[i, 2]);
        row -= centre;
        auto rotated_row = rot.rotate(row);
        rotated_row += centre;
        points(i, 0) = rotated_row[0];
        points(i, 1) = rotated_row[1];
        points(i, 2) = rotated_row[2];
    }
}

void Grid::operator*=(const double &_scale){
    points *= _scale;
    u *= _scale;
    v *= _scale;
}

void Grid::operator/=(const double &_scale){
    points /= _scale;
    u /= _scale;
    v /= _scale;
}

void Grid::operator+=(const Vec3& offset){
    points += offset.toNd(points.shape().rows);
}

void Grid::operator-=(const Vec3& offset){
    points -= offset.toNd(points.shape().rows);
}