#include "triangle.hpp"

using namespace QOSM::Pose;
using namespace QOSM::Item;

Medium::Medium(const bool isPEC) : n(1.),kappa(0.), isMetallic(isPEC), isPEC(isPEC){
    Z_ = 2 * std::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
}

Medium::Medium(const std::complex<double> &ior_, const bool isMetallic)
 : n(ior_.real()), kappa(-ior_.imag()), isMetallic(isMetallic), isPEC(false){
    double Z0 = 2 * std::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    Z_ = Z0 / (n - kappa*nc::constants::j);
}

Medium::Medium(const std::pair<double, double> &ior, const bool isMetallic)
 : n(ior.first), kappa(ior.second), isMetallic(isMetallic), isPEC(false){
    double Z0 = 2 * std::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    Z_ = Z0 / (n - kappa*nc::constants::j);
}

void Medium::setWithPermittivity(const std::complex<double> &epsilon_r_){
    const auto n_ = std::sqrt(epsilon_r_);
    n = n_.real();
    kappa = -n_.imag();
    double Z0 = 2 * std::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    Z_ = Z0 / n_;
}

void Medium::setWithPermittivity(const double &er, const double &ei){
    const auto n_ = std::sqrt(er - nc::constants::j * ei);
    n = n_.real();
    kappa = -n_.imag();
    double Z0 = 2 * std::sqrt(nc::constants::pi * 1e-7 / 8.854187817e-12);
    Z_ = Z0 / n_;
}

void Medium::setWithConductivity(const double &epsr, const double &conductivity, const double &freq){
    const double epsr_im = conductivity / (8.854187817e-12 * 2.*nc::constants::pi * freq);
    setWithPermittivity(epsr, epsr_im);
}

std::pair<double, double> Medium::get_ior(){
    return {n, kappa};
}

Dioptre::Dioptre(bool is_pec) : n1(1.), n2(1.), kappa1(0.), kappa2(0.), isMetallic(is_pec), isPEC(is_pec){}

Dioptre::Dioptre(const Medium &external_medium, const Medium &internal_medium) 
: n1(external_medium.n), n2(internal_medium.n), kappa1(external_medium.kappa), kappa2(internal_medium.kappa), isMetallic(internal_medium.isMetallic), isPEC(internal_medium.isPEC){}

Dioptre::Dioptre(const std::pair<double, double> &ior1, const std::pair<double, double> &ior2, const bool isMetallic)
 : n1(ior1.first), n2(ior2.first), kappa1(ior1.second), kappa2(ior2.second), isMetallic(isMetallic), isPEC(isPEC){}


Triangle::Triangle( const std::array<Vec3, 3> &vertices, 
                    const std::array<Vec3, 3> &normals,
                    const std::array<Vec3, 2> &edges, 
                    const std::array<double, 3> &curvatures, 
                    const Dioptre &dioptre, 
                    const int &idTriangle) : dioptre(dioptre), idTriangle(idTriangle){

    bounces[0] = -1;
    bounces[1] = -1;

    // set vertex position
    v0 = vertices[0];
    v1 = vertices[1];
    v2 = vertices[2];

    // set normal vector for each vertex
    n0 = normals[0];
    n1 = normals[1];
    n2 = normals[2];

    // set curvature for each vertex
    c0 = curvatures[0];
    c1 = curvatures[1];
    c2 = curvatures[2];
    
    // set edges
    edge1 = edges[0];
    edge2 = edges[1];

    // set triangle normal (at barycentre)
    n = (n1 + n2) * .5;
    n.normalise();
    
    // set triangle barycentre
    b = (v0 + v1 + v2) / 3.;
    
    // set dioptre curvature (at barycentre)
    c = (c1 + c2) * 0.5;
}

/**
 * Check if a ray intersects the triangle.
 *
 * @param ray The ray as a tuple (origin, direction).
 * @return true if an intersection has been found, false otherwise
 */

bool Triangle::intersect(const QOSM::Propagation::Ray &ray, double &t, double &u, double &v) const{
    constexpr float epsilon = 1e-8; //std::numeric_limits<float>::epsilon();
    const Vec3 &ray_ori = ray.ori;
    const Vec3 &ray_dir = ray.dir;

    const Vec3 h = ray_dir.cross(edge2);
    const double det = edge1.dot(h);

    // Check if the ray is parallel to the triangle surface
    if (std::fabs(det) < epsilon)
        return false;

    const double inv_det = 1.0 / det;
    const Vec3 s = ray_ori - v0;

    // Compute u
    u = s.dot(h) * inv_det;
    if (u < 0.0 || u > 1.0)
        return false;

    // Compute v
    Vec3 q = s.cross(edge1);
    v = ray_dir.dot(q) * inv_det;
    if (v < 0.0 || u + v > 1.0)
        return false;

    // Compute t
    t = edge2.dot(q) * inv_det;
    if (t < epsilon) // Intersection behind the ray origin
        return false;

    // Intersection found
    return true;
}
/*bool Triangle::intersect(const QOSM::Propagation::Ray &ray, double &t, double &u, double &v) const{
    constexpr float epsilon = std::numeric_limits<float>::epsilon();
    const Vec3 &ray_origin = ray.ori ;
    const Vec3 &ray_vector = ray.dir;

    Vec3 edge1 = v1 - v0;
    Vec3 edge2 = v2 - v0;
    Vec3 ray_cross_e2 = ray_vector.cross(edge2);
    float det = edge1.dot(ray_cross_e2);

    if (det > -epsilon && det < epsilon)
        return false;    // This ray is parallel to this triangle.

    float inv_det = 1.0 / det;
    Vec3 s = ray_origin - v0;
    u = inv_det * s.dot(ray_cross_e2);

    if (u < 0 || u > 1)
        return false;

    Vec3 s_cross_e1 = s.cross(edge1);
    v = inv_det * ray_vector.dot(s_cross_e1);

    if (v < 0 || u + v > 1)
        return false;

    // At this stage we can compute t to find out where the intersection point is on the line.
    t = inv_det * edge2.dot(s_cross_e1);

    if (t > epsilon) // ray intersection
    {
        return true;
    }
    else // This means that there is a line intersection but not a ray intersection.
        return false;
}*/

/**
 * Interpolate the normal for a point inside the triangle.
 *
 * @param u u-coord in barycentric coordinates
 * @param v v-coord in barycentric coordinates
 * @return The interpolated normal
 */
Vec3 Triangle::normal(double &u, double &v) const{
    return Vec3(n1 * u + n2 * v + n0 * (1 - u - v)).normalised();
}


/**
 * Interpolate the curvature for a point inside the triangle.
 *
 * @param normal The interpolated normal (result)
 * @param u u-coord in barycentric coordinates
 * @param v v-coord in barycentric coordinates
 * @return The interpolated curvature
 */
double Triangle::curvature(double &u, double &v) const{
    return (c1 * u + c2 * v + c0 * (1 - u - v));

}