#include <vsrc.hpp>
#include <omp.h>
using namespace QOSM::Propagation;
using namespace QOSM::EM::Source;

VirtualSource::VirtualSource(double freq_GHz) : Source(freq_GHz){
    freq = freq_GHz * 1e9;
    k0 =  2*pi * freq / c0;
    assumeEzIs0 = false;
}

void VirtualSource::init(std::vector<Beam> &list_beams){
    beams = list_beams;
}

VirtualSource VirtualSource::extends(const VirtualSource &vsrc){
    beams.insert(beams.end(), vsrc.beams.begin(), vsrc.beams.end());
    return *this;
}

FieldNx3 VirtualSource::eField(const VectorNx3 &r_pts_src) const{
    auto numPts = r_pts_src.numRows();
    const double *pts_ptr = r_pts_src.data();
   
    FieldNx3 e_field = FieldNx3(numPts);

    cplx* e_ptr = e_field.data();

    // Get the number of available threads
    int num_cores = 20;
    // Calculate the number of threads to use to achieve 85% CPU usage
    int target_num_cores = num_cores * 0.5;

    // Set the number of threads for OpenMP
    omp_set_num_threads(target_num_cores);

    int i;
    #pragma omp parallel for schedule(dynamic, 50)
    for(i=0; i<numPts*3; i+=3){
        Vec3 pts(pts_ptr[i], pts_ptr[i+1], pts_ptr[i+2]);
        e_ptr[i] = 0;
        e_ptr[i+1] = 0;
        e_ptr[i+2] = 0;

        for (const Beam &beam: beams){
            auto fields = beam.computeFields(pts, assumeEzIs0);
            e_ptr[i] += fields.first[0];
            e_ptr[i+1] += fields.first[1];
            e_ptr[i+2] += fields.first[2];
        }
    }
    return e_field;
}

FieldNx3 VirtualSource::hField(const VectorNx3 &r_pts_src) const{
    auto numPts = r_pts_src.numRows();
    const double *pts_ptr = r_pts_src.data();
    
    FieldNx3 h_field(numPts);

    cplx* h_ptr = h_field.data();

    // Get the number of available threads
    int num_cores = 20;
    // Calculate the number of threads to use to achieve 85% CPU usage
    int target_num_cores = num_cores * 0.85;

    // Set the number of threads for OpenMP
    omp_set_num_threads(target_num_cores);

    int i;
    #pragma omp parallel for schedule(dynamic, 50)
    for(i=0; i<numPts*3; i+=3){
        Vec3 pts(pts_ptr[i], pts_ptr[i+1], pts_ptr[i+2]);
        h_ptr[i] = 0;
        h_ptr[i+1] = 0;
        h_ptr[i+2] = 0;

        for (const Beam &beam: beams){
            auto fields = beam.computeFields(pts, assumeEzIs0);
            h_ptr[i] += fields.second[0];
            h_ptr[i+1] += fields.second[1];
            h_ptr[i+2] += fields.second[2];
        }
    }
    return h_field;
}

std::tuple<FieldNx3, FieldNx3, VectorNx3> VirtualSource::fields(const VectorNx3 &r_pts_src) const{
    return orientedFields(r_pts_src, Vec3());
}


std::tuple<FieldNx3, FieldNx3, VectorNx3> VirtualSource::orientedFields(const VectorNx3 &r_pts_src, const Vec3 &z_probe_src) const{
    auto numPts = r_pts_src.numRows();
    const double *pts_ptr = r_pts_src.data();
    
    FieldNx3 e_field(numPts);
    FieldNx3 h_field(numPts);
    VectorNx3 s_field(numPts);

    cplx* e_ptr = e_field.data();
    cplx* h_ptr = h_field.data();
    double* s_ptr = s_field.data();

    // Get the number of available threads
    int num_cores = 20;
    // Calculate the number of threads to use to achieve 85% CPU usage
    int target_num_cores = num_cores * 0.85;

    // Set the number of threads for OpenMP
    omp_set_num_threads(target_num_cores);

    int i;
    #pragma omp parallel for schedule(dynamic, 50)
    for(i=0; i<numPts*3; i+=3){
        Vec3 pts(pts_ptr[i], pts_ptr[i+1], pts_ptr[i+2]);

        e_ptr[i] = 0;
        e_ptr[i+1] = 0;
        e_ptr[i+2] = 0;

        h_ptr[i] = 0;
        h_ptr[i+1] = 0;
        h_ptr[i+2] = 0;

        for (const Beam &beam: beams){
            auto fields = beam.computeFields(pts, assumeEzIs0);

            e_ptr[i] += fields.first[0];
            e_ptr[i+1] += fields.first[1];
            e_ptr[i+2] += fields.first[2];

            h_ptr[i] += fields.second[0];
            h_ptr[i+1] += fields.second[1];
            h_ptr[i+2] += fields.second[2];
        }

        s_ptr[i] = (e_ptr[i+1] * std::conj(h_ptr[i+2]) - e_ptr[i+2] * std::conj(h_ptr[i+1])).real();
        s_ptr[i+1] =  (e_ptr[i+2] *  std::conj(h_ptr[i]) - e_ptr[i] * std::conj(h_ptr[i+2])).real();
        s_ptr[i+2] = (e_ptr[i] * std::conj(h_ptr[i+1]) - e_ptr[i+1] *  std::conj(h_ptr[i])).real();

        // always 0 is z_probe_src.norm() == 0 ==> compare <0 and not <= 0
        double dot = s_ptr[i] * z_probe_src[0] + s_ptr[i+1] * z_probe_src[1] + s_ptr[i+2] * z_probe_src[2];
        if (dot < 0){
            e_ptr[i] = e_ptr[i+1] = e_ptr[i+2] = h_ptr[i] = h_ptr[i+1] = h_ptr[i+2] = 0.;
            s_ptr[i] = s_ptr[i+1] = s_ptr[i+2] = 0.;
        }
    }
    return {e_field, h_field, s_field};
}


VirtualSource VirtualSource::filter(const BeamType &type){
    VirtualSource output(freq * 1e-9);
    for (Beam &beam : beams){
        if (beam.beamType == type){
            output.beams.push_back(beam);
        }
    }
    return output;
}

std::pair<VirtualSource, VirtualSource> VirtualSource::split(){
    VirtualSource gen(freq * 1e-9);
    VirtualSource init(freq * 1e-9);
    for (Beam beam : beams){
        if (beam.beamType == BeamType::INITIAL){
            init.beams.push_back(beam);
        } else {
            gen.beams.push_back(beam);
        }
    }
    return {gen, init};
}
