#include <propagation.hpp>
#include <bvh.hpp>
#include <tsqueue.hpp>
#include <array>
#include <deque>
#include <algorithm>
#include <armadillo>
#include <omp.h>

namespace qem = QOSM::EM;
namespace qpropa = QOSM::Propagation;
namespace qitem = QOSM::Item;
namespace qpose = QOSM::Pose;
namespace qsrc = QOSM::EM::Source;

std::array<double, 4> dMatT(std::array<double, 4> M){
    return {M[0], M[2], M[1], M[3]};
}

std::array<cplx, 4> cMatT(std::array<cplx, 4> M){
    return {M[0], M[2], M[1], M[3]};
}

std::array<cplx, 4> dcMatMul(const std::array<double, 4> &A, const std::array<cplx, 4> &B){
    return {
        A[0]*B[0] + A[1]*B[2], A[0]*B[1] + A[1]*B[3],
        A[2]*B[0] + A[3]*B[2], A[2]*B[1] + A[3]*B[3]
    };
}
std::array<cplx, 4> cdMatMul(const std::array<cplx, 4> &A, const std::array<double, 4> &B){
    return {
        A[0]*B[0] + A[1]*B[2], A[0]*B[1] + A[1]*B[3],
        A[2]*B[0] + A[3]*B[2], A[2]*B[1] + A[3]*B[3]
    };
}

// 2D case
std::vector<qpropa::Beam> fieldExpansion(
    const FieldNx1 &E, 
    const VectorNx3 &S, 
    const VectorNx3 &src_r_pts,
    const double &freq, 
    const int& dim,
    const double &w0,
    const double pw_threshold
){    
    const auto pw = E.intensity(true).reshape(-1, 1);
    auto rowIndices = nc::arange<int>(0, E.numRows()).getByMask(pw > pw_threshold);

    const auto N = rowIndices.size();
    std::cout << "[gbe1] Performing GBE with " << N << " beams ... ";

    const auto _r = src_r_pts.at(rowIndices, nc::Slice(0, 3));
    const auto _E = E.at(rowIndices);
    const auto _S = S.at(rowIndices, nc::Slice(0, 3));

    std::vector<QOSM::Pose::Vec3> points(N);
    for (int n=0; n<N; n++){
        points[n] = Vec3(_r.at(n, 0), _r.at(n, 1), _r.at(n, 2));
    }

    const array2c pol = {1., 1.};
    const array2d ior = {1., 0.};
    const array2d z0_axis = {0., 0.};
    const array2d z_range = {-1e30, 1e30};
    const array2d w0_axis = {w0, w0};

    // E-Field vector at each pt-point on the sphere (b vector)
    // 1->N    => Es_x_glo
    // N+1->2N => Es_y_glo    
    arma::cx_dvec src_E_surf(N);
    src_E_surf.fill(0.);

    // set all beam frames
    std::vector<Frame> frames(N);
    std::vector<Vec3> src_x_pts(N);
    std::vector<Vec3> src_y_pts(N);
    std::vector<qpropa::Beam> beams(N);

    #pragma omp parallel for
    for (int p=0; p<N; p++){

        Vec3 pt_R_pt_src = points.at(p);
        Vec3 src_z_p = Vec3{_S.at(p, 0), _S.at(p, 1), _S.at(p, 2)}.normalised();
        Vec3 src_y_p = Vec3(0, 1, 0);
        Vec3 src_x_p = src_y_p.cross(src_z_p);
        frames.at(p) = Frame(src_x_p, src_y_p, src_z_p, pt_R_pt_src);

        // store Ey dot y_p
        src_E_surf(p) = E.at(p);
    }

    const auto beam = qpropa::Beam(
        freq, ior, pol, w0_axis, z0_axis, 
        1.0, 0.0, 0.0, z_range, {0, 0});
    

    // A matrix
    arma::cx_dmat src_u_gb(N, N);
    src_u_gb.fill(0.);
       
    // Compute GB in their beam's local reference frame (n)
    //  (src) -> (n) => q_loc_glo
    //  ie for the current point -> in all local frames
    int N2 = N * N;
    #pragma omp parallel for
    for (int i=0; i<N2; i++){
        int p = i / N;
        int n = i % N;
       
        // current point
        Vec3 r_src_p = points.at(p);  

        // write p-point w.r.t. the n-beam frame
        const auto FRAME_src_n = frames.at(n);
        const Vec3 n_r_p = FRAME_src_n.toLocal(r_src_p);

        // compute scalar function from n-beam (computed in n-frame)
        std::array<cplx, 4> Q_;
        const auto res = gaussianScalarFn(Q_, beam, n_r_p);

        // store u dot y_p
        src_u_gb(p, n) = res[0];
    }

    // resolve linear system
    arma::cx_dvec alpha;
    bool flag = arma::solve(alpha, src_u_gb, src_E_surf, arma::solve_opts::no_approx);

    // update each beam polarisation coefficients
    if (flag){
        for (int p=0; p<N; p++){
            beams[p] = qpropa::Beam(
                freq, ior, {alpha(p)*double(dim==0), alpha(p)*double(dim==1)}, w0_axis, z0_axis, 
                1.0, 0.0, 0.0, z_range, {0, 0}, frames.at(p), p);
        }
        std::cout << "Done." << std::endl;
    } else {
        std::cout << "No solution found." << std::endl;
    }

    return beams;
}



// 3D case
std::vector<qpropa::Beam> fieldExpansion(
    const FieldNx3 &E, 
    const VectorNx3 &S, 
    const VectorNx3 &src_r_pts,
    const double &freq, 
    const double &w0,
    const double pw_threshold
){    
    const auto pw = E.intensity(true).reshape(-1, 1);
    
    auto rowIndices = nc::arange<int>(0, E.numRows()).getByMask(pw > pw_threshold);

    const auto N = rowIndices.size();
    const auto _r = src_r_pts.at(rowIndices, nc::Slice(0, 3));
    const auto _E = E.at(rowIndices, nc::Slice(0, 3));
    const auto _S = S.at(rowIndices, nc::Slice(0, 3));

    std::cout << "[gbe2] Performing GBE with " << N << " beams ... ";
    std::vector<QOSM::Pose::Vec3> points(N);
    for (int n=0; n<N; n++){
        points[n] = Vec3(_r(n, 0), _r(n, 1), _r(n, 2));
    }

    const array2c pol = {1., 1.};
    const array2d ior = {1., 0.};
    const array2d z0_axis = {0., 0.};
    const array2d z_range = {-1e30, 1e30};
    const array2d w0_axis = {w0, w0};

    // E-Field vector at each pt-point on the sphere (b vector)
    // 1->N    => Es_x_glo
    // N+1->2N => Es_y_glo    
    arma::cx_dvec src_E_surf(2*N);
    src_E_surf.fill(0.);

    // set all beam frames
    std::vector<Frame> frames(N);
    std::vector<Vec3> src_x_pts(N);
    std::vector<Vec3> src_y_pts(N);
    std::vector<qpropa::Beam> beams(N);

    #pragma omp parallel for
    for (int p=0; p<N; p++){
        Vec3 pt_R_pt_src = points.at(p);
        Vec3 src_z_p = Vec3{_S.at(p, 0), _S.at(p, 1), _S.at(p, 2)}.normalised();
        Vec3 src_y_p = src_z_p.cross(Vec3(1, 0, 0));
        const double norm = src_y_p.normalise();
        Vec3 src_x_p;
        if (std::isnan(norm) || norm < 1e-15){
            src_x_p = {1., 0, 0};
            src_y_p = {0, -1., 0};
        } else 
            src_x_p = src_y_p.cross(src_z_p);
        frames.at(p) = Frame(src_x_p, src_y_p, src_z_p, pt_R_pt_src);
        src_x_pts.at(p) = src_x_p;
        src_y_pts.at(p) = src_y_p;

        const std::array<cplx, 3> src_E_p = {_E.at(p, 0), _E.at(p, 1), _E.at(p, 2)};
        // Es dot xp
        src_E_surf(p)   = src_E_p[0] * src_x_p[0] + src_E_p[1] * src_x_p[1] + src_E_p[2] * src_x_p[2];
        // Es dot yp
        src_E_surf(p+N) = src_E_p[0] * src_y_p[0] + src_E_p[1] * src_y_p[1] + src_E_p[2] * src_y_p[2];
    }

    const auto beam = qpropa::Beam(
        freq, ior, pol, w0_axis, z0_axis, 
        1.0, 0.0, 0.0, z_range, {0, 0});

    // A matrix
    arma::cx_dmat src_E_gb(2*N, 2*N);
    src_E_gb.fill(0.);
       
    // Compute GB in their beam's local reference frame (n)
    //  (src) -> (n) => q_loc_glo
    //  ie for the current point -> in all local frames
    int N2 = N * N;
    #pragma omp parallel for
    for (int i=0; i<N2; i++){
        int p = i / N;
        int n = i % N;
       
        // current reference beam frame (w.r.t. src frame)
        Vec3 r_src_p = points.at(p);  
        const Vec3 &src_x_p = src_x_pts.at(p);
        const Vec3 &src_y_p = src_y_pts.at(p);

        // write p-point w.r.t. the n-beam frame
        const auto FRAME_src_n = frames.at(n);
        const Vec3 n_r_p = FRAME_src_n.toLocal(r_src_p);
        const double &n_rx_p = n_r_p[0];
        const double &n_ry_p = n_r_p[1];

        // compute field from n-beam (computed in n-frame)
        std::array<cplx, 4> Q_;
        const auto res = gaussianScalarFn(Q_, beam, n_r_p);
        const cplx u_ = res[0];

        // write the field vector for each polarisation -> (xz) and (yz)
        // note: here Q12 and Q21 are both 0 since w0_x = w0_y => Qd = 0
        std::array<cplx, 3> n_Exz_p{
            u_,
            0., 
            -Q_[0] * n_rx_p * u_
        };
        std::array<cplx, 3> n_Eyz_p{
            0., 
            u_, 
            -Q_[3] * n_ry_p * u_
        };

        // write thos vector in src frame
        const auto src_Exz_p = FRAME_src_n.rot2Ref(n_Exz_p);
        const auto src_Eyz_p = FRAME_src_n.rot2Ref(n_Eyz_p);

        // dot xp
        const auto src_exz_x = src_Exz_p[0] * src_x_p[0] + src_Exz_p[1] * src_x_p[1] + src_Exz_p[2] * src_x_p[2];
        const auto src_eyz_x = src_Eyz_p[0] * src_x_p[0] + src_Eyz_p[1] * src_x_p[1] + src_Eyz_p[2] * src_x_p[2];

        // dot yp
        const auto src_exz_y = src_Exz_p[0] * src_y_p[0] + src_Exz_p[1] * src_y_p[1] + src_Exz_p[2] * src_y_p[2];
        const auto src_eyz_y = src_Eyz_p[0] * src_y_p[0] + src_Eyz_p[1] * src_y_p[1] + src_Eyz_p[2] * src_y_p[2];

        // store the dot products in the matrix (A)
        src_E_gb(p, n) = src_exz_x;
        src_E_gb(p, n + N) = src_eyz_x;
        src_E_gb(p + N, n) = src_exz_y;
        src_E_gb(p + N, n + N) = src_eyz_y;
    }

    // resolve linear system
    // alpha x: alpha(i)
    // alpha y; alpha(i+N)
    arma::cx_dvec alpha;
    bool flag = arma::solve(alpha, src_E_gb, src_E_surf); //, arma::solve_opts::no_approx);

    // update each beam polarisation coefficients
    if (flag){
        for (int p=0; p<N; p++){
            beams[p] = qpropa::Beam(
                freq, ior, {alpha(p), alpha(p + N)}, w0_axis, z0_axis, 
                1.0, 0.0, 0.0, z_range, {0, 0}, frames.at(p), p);
        }
        std::cout << "Done." << std::endl;
    } else {
        std::cout << "No solution found." << std::endl;
    }

    return beams;
}


std::pair<bool, bool> computeRT(
    qpropa::Beam &beamI0, 
    qpropa::Beam &beamR, 
    qpropa::Beam &beamT, 
    qitem::Dioptre &dioptre, 
    const double &dist, 
    qpose::Vec3 normal, 
    const double &curvature, 
    const double &thresh,
    const int idTriangle
){
    std::pair<bool, bool> beam_created = {false, false};

    if (beamI0.zRange[1] < dist){
        return beam_created;
    }

    beamI0.zRange[1] = dist;

    if (beamI0.bouncesLeft[0] < 1 && beamI0.bouncesLeft[1] < 1){
        return beam_created;
    }

    double k0 = beamI0.k0;

    auto r_i_src = beamI0.frame.ori + beamI0.frame.z * dist;
    double curv, n1, kappa1, n2, kappa2;

    bool isMetallic = dioptre.isMetallic;
    bool isPEC = dioptre.isPEC;
    const auto n0 = normal;

    // determine the sides of the intersection
    // the normal must point towards the entry side
    // then: k_i dot n must be negative
    if (normal.dot(beamI0.frame.z) < 0){
        // nominal condition
        curv = -curvature;
        n1 = dioptre.n1;
        n2 = dioptre.n2;
        kappa1 = dioptre.kappa1;
        kappa2 = dioptre.kappa2;
    } else {
        // reversed condition. We must flip the normal and use the opposite of the curvature
        normal = normal * -1;
        curv = curvature;
        n1 = dioptre.n2;
        n2 = dioptre.n1;
        kappa1 = dioptre.kappa2;
        kappa2 = dioptre.kappa1;
    }

    cplx j(0., 1.);

    // ratio of propagation constants
    // gamma1 / gamma2 = (n1 - 1j*kappa1)/(n2 - 1j*kappa2)
    double den = n2 * n2 + kappa2 * kappa2;
    cplx gamma_1o2 = n1 / n2;

    // Fresnel coefficients uses n + j*kappa convention
    cplx n1_ = n1 + kappa1 * j;
    cplx n2_ = n2 + kappa2 * j;
    cplx gamma_1o2_ = n1_ / n2_;

    double cosTheta1 = normal.dot(beamI0.frame.z * -1.);
    cplx sinTheta1Sq_ = (1 - cosTheta1*cosTheta1);
    cplx sinTheta2Sq_ = gamma_1o2 * gamma_1o2 * sinTheta1Sq_;

    cplx cosTheta2sq_ = 1. - sinTheta2Sq_;
    cplx cosTheta2_ = std::sqrt(cosTheta2sq_);
    double cosTheta2 = cosTheta2_.real();

    // case of evanescent transmitted wave
    // No propagation means than there is no real part of the effective transmitted propagation constant
    bool int_refl = cosTheta2 < 1e-5;

    // Complex Fresnel coefficients
    array2c r, t;
    if (isPEC){
        r = {1., -1.};
        t = {0., 0.};
    } else {
        r = {
            (n2_ * cosTheta1 - n1_ * cosTheta2_) / (n2_ * cosTheta1 + n1_ * cosTheta2_),  // P-Polarisation (TM) --> x-axis
            (n1_ * cosTheta1 - n2_ * cosTheta2_) / (n1_ * cosTheta1 + n2_ * cosTheta2_)   // S-Polarisation (TE) --> y-axis
        }; 
        if (isMetallic || int_refl)
            t = {0., 0.};
        else{
            t = {
                (r[0] + 1.) * n1_/n2_,  // P-Polarisation (TM)   
                r[1] + 1.,              // S-Polarisation (TE)
            };
        }
    }
    
    // incident ray direction
    Vec3 zi = beamI0.frame.z.normalised();

    // reflected ray direction
    Vec3 zr = (zi + normal * 2 * cosTheta1).normalised();

    // incident ray fixed frame
    Frame frameI0 = beamI0.frame;

    // local incident ray fixed frame
    Vec3 xi, yi;
    if (zi.cross(normal).norm2() < 1e-10){
        // normal incidence
        yi = frameI0.y;
        xi = frameI0.x;;
    } else {
        xi = zi.cross(normal.cross(zi).normalised());
        yi = zi.cross(xi).normalised();
    }
    
    Frame frameI(xi, yi, zi, r_i_src);

    // get tangent space (force v to be yi)
    Vec3 v = yi;
    Vec3 u = v.cross(normal*-1);

    // from incident frame to local incident frame (SP polarisations)
    // v_i = K.v_i0
    std::array<double, 4> K = {
        frameI.x.dot(frameI0.x), frameI.x.dot(frameI0.y), 
        frameI.y.dot(frameI0.x), frameI.y.dot(frameI0.y),
    };

    // compute the polarisation of the local incident beam
    //   polarisation[0] = TM
    //   polarisation[1] = TE
    array2c polarisation = {
        beamI0.alpha_[0] * K[0] + beamI0.alpha_[1] * K[1],
        beamI0.alpha_[0] * K[2] + beamI0.alpha_[1] * K[3]
    };
    // initialisation parameters
    // Incident beam curvature matrix at intersection point
    Vec3 point = {0., 0., dist};
    std::array<cplx, 4> Qi0;
    auto res = gaussianScalarFn(Qi0, beamI0, point);
    cplx amplI = res[1];
    cplx phaseI = res[2];

    // from incident frame to tangent-space
    // Q_ts = K.Q_i0.K'
    std::array<double, 4> Ki ={
        frameI0.x.dot(u), frameI0.x.dot(v),
        frameI0.y.dot(u), frameI0.y.dot(v)
    };

    // write incident beam curvature matrix in Tangent Space
    std::array<cplx, 4> Qi_ts = dcMatMul(dMatT(Ki), cdMatMul(Qi0, Ki));
    const double dot1 = normal.dot(zi);

    // Reflected ray
    array2c fresnelR = {polarisation[0] * r[0], polarisation[1] * r[1]};
    double pwR = std::pow(std::abs(fresnelR[0]), 2) + std::pow(std::abs(fresnelR[1]), 2);
    pwR /= beamI0.initialPower;
    if (pwR > thresh && beamI0.bouncesLeft[0] > 0){
        // set the reflected ray fixed frame
        Vec3 xr, yr;
        if (zr.cross(normal).norm2() < 1e-8){
            // normal incidence, keep same S-polarisation (y)
            yr = yi;
            xr = yr.cross(zr).normalised();
        } else {
            xr = zr.cross(normal.cross(zr).normalised());
            yr = zr.cross(xr).normalised();
        }
        Frame frameR(xr, yr, zr, r_i_src /*+ zr*epsilon*/);

        // from tangent-space to reflected beam frame
        // Qr_r = Kr.Qr_ts.Kr'
        std::array<double, 4> Kr ={
            frameR.x.dot(u), frameR.x.dot(v),
            frameR.y.dot(u), frameR.y.dot(v)
        };
        double detKr = 1. / (Kr[0] * Kr[3] - Kr[1] * Kr[2]);
        std::array<double, 4> invKr = {
            detKr * Kr[3], - detKr * Kr[1],
            -detKr * Kr[2], detKr * Kr[0]
        };  
        const auto KrT = dMatT(Kr);
        double detKrT = 1 / (KrT[0] * KrT[3] - KrT[2] * KrT[1]);
        std::array<double, 4> invKtT = {
              detKrT * KrT[3],   - detKrT * KrT[1],
            - detKrT * KrT[2],     detKrT * KrT[0]
        };
        const double dot2 = normal.dot(zr);

        auto curv_term = curv * (dot1 - dot2);
        std::array<cplx, 4> Qr_ts = {
            Qi_ts[0] - curv_term,   Qi_ts[1],
            Qi_ts[2],               Qi_ts[3] - curv_term
        };

        // no need to take account of curvature as n.dot(zi) - n.dot(zr) is always 0
        std::array<cplx, 4> Qr = dcMatMul(dMatT(invKr), cdMatMul(Qr_ts, invKr));
        
        beamR.frame = frameR;
        beamR.k0 = beamI0.k0;
        beamR.fresnel_ = r;
        beamR.alpha_ = fresnelR;
        beamR.bouncesLeft = {beamI0.bouncesLeft[0] - 1, beamI0.bouncesLeft[1]};
        beamR.zRange = {0., 1e30};
        beamR.beamType = BeamType::REFLECTED;
        beamR.initialPower = beamI0.initialPower;
        // reflection in the same media than incident beam
        beamR.n = beamI0.n;
        beamR.kappa = beamI0.kappa;
        beamR.Z_ = beamI0.Z_; 
        beamT.id_triangle = idTriangle;
        beamR.trck_id = beamI0.trck_id;

        QOSM::Propagation::initBeam(beamR, Qr, amplI, phaseI);
        beam_created.first = true;
    }
    
    // Transmitted ray
    double pwT = 1. - pwR;
    if (!isMetallic && !int_refl && beamI0.bouncesLeft[1] > 0 && pwT > thresh){
        const auto ratioN = n1 / n2;
        
        //  set the transmitted ray fixed frame
        Vec3 zt, xt, yt;
        if (zi.cross(normal).norm2() < 1e-8){
            // normal incidence
            xt = xi;
            yt = yi;
            zt = zi;
        } else {
            zt = zi * ratioN + normal * (ratioN * cosTheta1 - cosTheta2);  // transmitted ray direction
            xt = (zt*-1).cross(normal.cross(zt*-1)).normalised();
            yt = zt.cross(xt);
        }
        Frame frameT(xt, yt, zt, r_i_src/* + zt*epsilon*/);

        // from tangent-space to transmitted beam frame
        // Qt_t = Kt.Qt_ts.Kt'  
        std::array<double, 4> Kt = {
            frameT.x.dot(u), frameT.x.dot(v),
            frameT.y.dot(u), frameT.y.dot(v)
        };
        double detKt = 1 / (Kt[0] * Kt[3] - Kt[2] * Kt[1]);
        std::array<double, 4> invKt = {
              detKt * Kt[3],   - detKt * Kt[1],
            - detKt * Kt[2],     detKt * Kt[0]
        };
        const auto KtT = dMatT(Kt);
        double detKtT = 1 / (KtT[0] * KtT[3] - KtT[2] * KtT[1]);
        std::array<double, 4> invKtT = {
              detKtT * KtT[3],   - detKtT * KtT[1],
            - detKtT * KtT[2],     detKtT * KtT[0]
        };
        const double dot2 = normal.dot(zt);

        auto curv_term = curv * (dot1 - n2/n1 * dot2);
        std::array<cplx, 4> Qt_ts = {
            gamma_1o2_ * (Qi_ts[0] + curv_term),    gamma_1o2_ * Qi_ts[1],
            gamma_1o2_ * Qi_ts[2],                  gamma_1o2_ * (Qi_ts[3] + curv_term)
        };
        std::array<cplx, 4> Qt = dcMatMul(invKtT, cdMatMul(Qt_ts, invKt));

        // transmitted beam polarisation
        array2c fresnelT = {polarisation[0] * t[0], polarisation[1] * t[1]};
        
        beamT.frame = frameT;
        beamT.k0 = beamI0.k0;
        beamT.alpha_ = fresnelT;
        beamT.fresnel_ = t;
        beamT.bouncesLeft = {beamI0.bouncesLeft[0], beamI0.bouncesLeft[1] - 1};
        beamT.zRange = {0., 1e30};
        beamT.beamType = BeamType::TRANSMITTED;
        beamT.initialPower = beamI0.initialPower;
        beamT.n = n2;
        beamT.kappa = kappa2;
        beamT.Z_ = Z0 / n2_;
        beamT.trck_id = beamI0.trck_id;
        beamT.id_triangle = idTriangle;
        beam_created.second = true;

        QOSM::Propagation::initBeam(beamT, Qt, amplI, phaseI);
        
        if (beamT.q1_.imag() < 0 || beamT.q2_.imag() < 0){
            // evanescent ray, dump it
            beam_created.second = false;
        }
    } 
    return beam_created;
}



std::deque<qpropa::Beam> beamTracing(
    std::vector<std::shared_ptr<qitem::Triangle>>& geometry, 
    std::vector<qpropa::Beam>& beams, 
    const array2i& numBouces, 
    const double& thresh, 
    const bool &incrementBouncing
){
    std::vector<qpropa::beam_ptr> beams_ptr;

    if (geometry.empty()){
        throw std::runtime_error("Geometry is empty") ;
    }
    
    std::cout << "Building BVH for " << geometry.size() << " triangles ...";
    auto bvhRoot = build_bvh(geometry);
    std::cout << " Done. " << std::endl;

    TSQueue<qpropa::Beam> output;
    int i;
    size_t N = beams.size();

    // Get the number of available threads
    int num_cores = 20;
    // Calculate the number of threads to use to achieve 95% CPU usage
    int target_num_cores = num_cores * 0.95;
    std::cout << "Performing GBT (bounces: " << numBouces[0] << ", " << numBouces[1] << ") ... ";

    // Set the number of threads for OpenMP
    omp_set_num_threads(target_num_cores);

    #pragma omp parallel for schedule(dynamic)
    for(i=0; i < N; i++){
        Beam &beam = beams.at(i);
        if (incrementBouncing){
            beam.bouncesLeft[0] += numBouces[0];
            beam.bouncesLeft[1] += numBouces[1];
        } else {
            beam.bouncesLeft = numBouces;
        }
        
        beam.beamType = BeamType::INITIAL;

        std::deque<qpropa::Beam> processQueue;
        processQueue.push_back(beam);

        while(!processQueue.empty()){
            // get next beam from process list
            Beam beamI = processQueue.front();
            processQueue.pop_front();

            // if (beamI.bouncesLeft[0] == 0 && beamI.bouncesLeft[1] == 0) continue;

            // check intersections
            Ray ray = beamI.getRay();
            Intersection intersection = ray_intersect_bvh(ray, bvhRoot, 1);

            // compute reflection and transmission
            if (intersection.valid){
                const int idTriangle = intersection.triangle->idTriangle;
                Beam beamR, beamT;
                Vec3 normal = intersection.triangle->normal(intersection.u, intersection.v);
                double curv = intersection.triangle->curvature(intersection.u, intersection.v);
                auto created = computeRT(beamI, beamR, beamT, intersection.triangle->dioptre, intersection.t, normal, curv, thresh, idTriangle);
                if (created.second)
                    processQueue.push_back(beamT);
                
                if (created.first)
                    processQueue.push_back(beamR);            

            }
            
            // add this beam to the result queue
            output.push(beamI);
        }
    }

    std::cout << "Done (beams: " <<  beams.size() << " -> " << output.size() << ")." << std::endl;
    return output.queue();
}

std::deque<qpropa::Beam> beamTracing(
    std::vector<qitem::Triangle>& geometry, 
    std::vector<qpropa::Beam>& beams, 
    const array2i& numBouces, 
    const double& thresh, 
    const bool &incrementBouncing
){
    std::vector<qitem::triangle_ptr> triangles_ptr;
    std::vector<qpropa::beam_ptr> beams_ptr;

    if (geometry.empty()){
        throw std::runtime_error("Geometry is empty") ;
    }
    
    for (qitem::Triangle &triangle : geometry){
        triangles_ptr.push_back(std::make_shared<qitem::Triangle>(std::move(triangle)));
    }

    return beamTracing(triangles_ptr, beams, numBouces, thresh, incrementBouncing);
}


std::deque<qpropa::Beam> surfaceBeamTracing(
    std::vector<qpose::Vec3>& normals, 
    std::vector<double>& curvatures, 
    Dioptre& dioptre, 
    std::vector<qpropa::Beam>& beams, 
    const double &thresh
){
    TSQueue<qpropa::Beam> output;
    int i;
    int N = normals.size();
    if (normals.size() != curvatures.size() || normals.size() != beams.size() || curvatures.size() != beams.size()){
        THROW_RUNTIME_ERROR("Inconsistent input size");
    }

    for(i=0; i<N; i++){
        Beam beamR, beamT;
        Beam &beamI = beams.at(i);
        auto created = computeRT(beamI, beamR, beamT, dioptre, 0., normals.at(i), curvatures.at(i), thresh, -1);
        output.push(beamI);
        
        if (created.second)
            output.push(beamT);
        
        if (created.first)
            output.push(beamR); 
    }
    return output.queue();
}

double computeCollectedPower(const std::shared_ptr<qsrc::Source> &src, 
                             const qpose::VectorNx3 &aperturePoints, 
                             const double &dudv, 
                             const qpose::Vec3 &z_probe_src){

    auto res = src->fields(aperturePoints);
    auto &P = std::get<2>(res);
    auto power = P.dot(z_probe_src.toNd(1));

    return nc::trapz(power, dudv)[0];    
}
